
#include <string.h>
#include <stdlib.h>
#include <ctype.h>


#include "parser.h"
 typedef struct {
	char *str;
	parser_sink_t *sink;
} parser_t;
 static  void output_bullet_pbreak(parser_t *p,int d) ; static  void output_paragraph_break(parser_t *p) ; static  void output_link(parser_t *p,int kind, char *s,char *e) ; static  void output_text(parser_t *p,int style,char *s,char *e) ; static  void output_preformatted(parser_t *p,char *s,char *e) ; static  void output_header(parser_t *p,int level,char *s,char *e) ; static  void parse_bullet(parser_t *p) ; static  void paragraph_break(parser_t *p) ; static  void parse_header(parser_t *p) ; static  void parse_preformatted(parser_t *p) ; static  void parse_text(parser_t *p) ; static  void parse_inline_attribs(parser_t *p) ; static  void parse_normal_text(parser_t *p) ;







 static  void output_bullet_pbreak(parser_t *p,int d) { (*p->sink->bullet)(p->sink, d); }
 static  void output_paragraph_break(parser_t *p) { (*p->sink->pbreak)(p->sink); }
 static  void output_link(parser_t *p,int kind, char *s,char *e) {
	if (e!=s)
		(*p->sink->link)(p->sink, kind, s, e-s);
}
 static  void output_text(parser_t *p,int style,char *s,char *e) {
	if (e!=s)
		(*p->sink->text)(p->sink, style,s, e-s);
}
 static  void output_preformatted(parser_t *p,char *s,char *e) {
	if (e!=s)
		(*p->sink->formatted)(p->sink, s, e-s);
}
 static  void output_header(parser_t *p,int level,char *s,char *e) {
	if (e!=s)
		(*p->sink->header)(p->sink, level,s, e-s);
}








 void parse_doc(char *str,parser_sink_t *sink) {
	parser_t parser;
	parser_t *p= &parser;
	int own_buffer;
	parser.sink= sink;

	 {
	int L;
	int i;
	own_buffer= 0;
	L= strlen(str);
	for(i=0;i<L;i++)
		if (str[i]=='\r') {
			own_buffer=1;
			break;
		}
	if (own_buffer) {
		int j;
		char *nb= malloc(L+1);
		j= 0;
		for(i=0;i<=L;i++)
			if (str[i]!='\r') nb[j++]= str[i];
		str= nb;
	}
}
			



	parser.str= str;

	while(parser.str[0]) {
		switch(parser.str[0]) {
			case '\n': paragraph_break(p); break;
			case '=': parse_header(p); break;
			case '+': parse_bullet(p); break;
			case ' ': parse_preformatted(p); break;
			case '\t': parse_preformatted(p); break;
			default: parse_text(p); 
		}
	}
	if (own_buffer)
		free(str);
}

 static  void parse_bullet(parser_t *p) {
	int level=0;
	while(p->str[0]=='+') {
		level++;
		p->str++;
	}
	if (p->str[0]=='\n')
		return ;

	while(1) {
		switch(p->str[0]) {
		case '\n': p->str++;
		case 0: return ;
		default:
			if (isspace(p->str[0])) 
				p->str++;
			else {
				output_bullet_pbreak(p,level);
				parse_text(p);
				return ;
			}
		}
	}
}


 static  void paragraph_break(parser_t *p) {
	output_paragraph_break(p);
	p->str++;
}


 static  void parse_header(parser_t *p) {
	char *s, *e;
	int level=0;
	int state= 'I';
	while(state!='X') {
		switch(state) {
		case 'I': switch(p->str[0]) {
			case '=': level++; p->str++; break;
			case '\n': p->str++; return ;
			case 0: return ;
			default: state= 'C'; s= p->str; p->str++;
			} break;
		case 'C': switch(p->str[0]) {
			case '=': e= p->str; state= 'S'; break;
			case '\n': e= p->str; state= 'N'; break;
			case 0: e= p->str; state= 'X'; break;
			default: p->str++;
			} break;
		case 'S': switch(p->str[0]) {
			case '=': p->str++; break;
			case 0: state= 'X'; break;
			case '\n': state= 'N'; break;
			default: p->str++;
			} break;
		case 'N': p->str++; state= 'X'; break;
		}
	}
	if (level>4) level= 4;
	output_header(p, 5-level, s,e);
}

 static  void parse_preformatted(parser_t *p) {
	char *s,*e;
	int done= 0;
	s= p->str;
	do {
		switch(p->str[0]) {
		case '\n': p->str++; e= p->str; done= 1; break;
		case 0: e= p->str; done= 1; break;
		default: p->str++;
		}
	} while(!done);
	output_preformatted(p, s,e);
}

 static  void parse_text(parser_t *p) {
	int done= 0;
	do {
		switch(p->str[0]) {
		
	case 'L': case 'X': case 'W':


		
	case 'R': case 'B': case 'I':
	case 'G': case 'C':


			parse_inline_attribs(p);
			break;
		case '\n':
			p->str++;
		case 0:
			done= 1;
			break;
		default: 
			parse_normal_text(p);
		}
	} while(!done);
}

 static  void parse_inline_attribs(parser_t *p) {
	int c;
	int b;
	int done;
	char *s,*e;
	c= p->str[0];
	p->str++;
	if (p->str[0]!='{' && p->str[0]!='<') {
		output_text(p, 'D', p->str-1, p->str);
		return ;
	}
	b= p->str[0];
	p->str++;
	s= p->str;
	done= 0;
	do {
		switch(p->str[0]) {
		case 0:
			e= p->str;
			done= 1;
			break;
		case '>': 
			if (b=='<') {
				e= p->str;
				p->str++;
				done= 1;
			} else p->str++;
			break;
		case '}':
			if (b=='{') {
				e= p->str;
				p->str++;
				done= 1;
				break;
			}
		default:
			p->str++;
		}
	} while(!done);
	switch(c) {
	
	case 'L': case 'X': case 'W':


		output_link(p,c,s,e);
		break;
	
	case 'R': case 'B': case 'I':
	case 'G': case 'C':


		output_text(p, c,s,e);
		break;
	}
}

/* this function doesnt eat the newline so that 
the parent can look at it. */
 static  void parse_normal_text(parser_t *p) {
	char *s,*e;
	int done= 0;
	s= p->str;
	do {
		switch(p->str[0]) {
		case '\n':
		case 0:
		
	case 'L': case 'X': case 'W':


		
	case 'R': case 'B': case 'I':
	case 'G': case 'C':


			e= p->str; 
			if (*e=='\n') e++;
			done= 1; 
			break;
		default:
			p->str++;
		}
	} while(!done);
	output_text(p, 'D', s,e);
}

/* ok. so, 
	\n\n -> paragraph break
	\n= -> header
	\n+ -> list
	\n\t -> preformatted
	\nSPC -> preformatted
	\n.  -> normal text

	ok. so just assume that we enter this function only
	at the beginning of a line. cool.
	
	if a piece of styled text spans multiple lines, then pass those
	lines within the styled text and  then come back here. ok.

	we assume that the newline has already been eaten.

	we need a little bit of buffering at the output so that it makes
	sense. ie. outputting two paragraph breaks in a line does not
	make sense. neither does a paragraph break followed by a
	heading. (the paragraph break should be removed, then)
*/







