

#include <xincs.h>
#include <fxver.h>
#include <fxdefs.h>
#include <FXHash.h>
#include <FXThread.h>
#include <FXStream.h>
#include <FXString.h>
#include <FXRectangle.h>
#include <FXRegistry.h>
#include <FXApp.h>
#include <FXDCWindow.h>
#include <FXFont.h>


#include "xtview.h"


using namespace FX;


enum { st_def, st_red, st_green, st_bold, st_italic,
	st_code, st_link,
	st_head1, st_head2, st_head3, st_head4,
	st_pfmt , st_last };

struct xtv_style {
	FXFont *font;
	FXColor fgcolor;
};


 static  FXColor rgb24toFXColor(unsigned C) ; static  int docstyle_to_vstyle(int dstyle) ; static  void create_styles() ; static  void font_metrics(layout_metrics_t *m,int style,wchar_t chr,
		int *fascent, int *fdescent, int *width) ; static  int bullet_indent(layout_metrics_t *m,int level) ; static  int text_style(layout_metrics_t *m,int kind) ; static  int header_style(layout_metrics_t *m,int level) ; static  int tabstop(layout_metrics_t *m,int x) ; static  int find_element(int x,int y,text_part_t *parts, int nparts) ; static  void destroy_parts(int N,text_part_t *P) ; static  void min2first(int *A,int B) ; static  void max2first(int *A,int B) ; static  int rect_sub_inside(int *a, int *o) ;
 static  doc_style_t od_styles[st_last]; static  int width_of_M; static  int width_of_SPC;


FXDEFMAP(xtview_t) xtviewMap[]={
	FXMAPFUNC(SEL_PAINT,0,xtview_t::onPaint),
	FXMAPFUNC(SEL_LEFTBUTTONPRESS,0, xtview_t::onLeftBtnPress),
};


FXIMPLEMENT(xtview_t,FXScrollArea,xtviewMap,ARRAYNUMBER(xtviewMap))


xtview_t::xtview_t(){ flags|=FLAG_ENABLED|FLAG_SHOWN; }


static struct xtv_style styles[st_last];
static int styles_created;


xtview_t::xtview_t
		(FXComposite* p, FXuint opts,
		FXint x,FXint y,FXint w,FXint h):
			FXScrollArea(p,opts | VSCROLLER_ALWAYS ,x,y,w,h){
	flags|=FLAG_ENABLED;
	buffer= 0;
	parts= 0;
	nparts= 0;
	link_handler=0;

	cont_x= cont_y= cont_width= cont_height= 0;
}



 static  FXColor rgb24toFXColor(unsigned C) {
	unsigned R= (C>>16u)&0xffu;
	unsigned G= (C>>8u)&0xffu;
	unsigned B= C&0xffu;
	return FXRGB(R,G,B);
}


bool xtview_t::canFocus() const { return true; }

int xtview_t::getDefaultWidth() { return 0; }
int xtview_t::getDefaultHeight() { return 0; }

int xtview_t::getContentHeight() {
	if (!nparts) return 200;
	return cont_y+cont_height;
}

int xtview_t::getContentWidth() {
	if (!nparts) return 200;
	return cont_x+cont_width;
}


void xtview_t::create() {
	FXScrollArea::create();
	create_styles();
}

void xtview_t::detach() {
	FXScrollArea::detach();
}

long xtview_t::onLeftBtnPress(FXObject* obj,FXSelector sel,void* ptr){
	FXEvent *event= (FXEvent*) ptr;
	if (!event) 
		return 1;
	int evX= event->win_x-pos_x;
	int evY= event->win_y-pos_y;
	int p= find_element(evX, evY, parts, nparts);
	if (p!=-1) {
		if (parts[p].opcode=='T') 
			if (parts[p].d.text.aux && link_handler) 
				(*link_handler)(parts[p].d.text.aux);
	}
	return 1;
}


 static  int docstyle_to_vstyle(int dstyle) {
	switch(dstyle) {
	case 'D': return st_def; 
	case 'R': return st_red;
	case 'G': return st_green;
	case 'B': return st_bold;
	case 'I': return st_italic;
	case 'C': return st_code;
	case 'L': return st_link;
	case '1': return st_head1;
	case '2': return st_head2;
	case '3': return st_head3;
	case '4': return st_head4;
	case 'F': return st_pfmt;
	}
	return st_def;
}


 void xtview_set_styles(FX::FXApp *app,doc_style_t *doc_styles) {
	int i;
	
	for(i=0;doc_styles[i].id;i++) {
		int st= docstyle_to_vstyle(doc_styles[i].id);
		od_styles[st]= doc_styles[i];
	}
	for(i=0;i<st_last;i++) {
		if (!od_styles[i].font_name)
			od_styles[i].font_name=  od_styles[st_def].font_name;
		if (!od_styles[i].size)
			od_styles[i].size= od_styles[st_def].size;
	}
	for(i=0;i<st_last;i++) {
		int weight;
		int slant;
		if (od_styles[i].bold)
			weight= FXFont::Bold;
		else
			weight= FXFont::Normal;
		if (od_styles[i].italic)
			slant= FXFont::Italic;
		else
			slant= FXFont::Straight;
		styles[i].font= new FXFont(app,od_styles[i].font_name,
			od_styles[i].size, weight, slant);
		styles[i].fgcolor= rgb24toFXColor(od_styles[i].fgcolor);
	}
}

 static  void create_styles() {
	int i;
	if (styles_created)
		return ;
	styles_created= 1;
	for(i=0;i<st_last;i++)
		styles[i].font->create();
}

void xtview_t::set_text(char *buf) {
	if (parts)
		destroy_parts(nparts,parts);
	nparts= 0;
	parts= 0;
	if (buffer)
		free(buffer);
	buffer= strdup(buf);
	recalc();
	layout();
	update();
}

void xtview_t::layout_buffer() {
	
	layout_metrics_t metrics;
	{ layout_metrics_t *m= &metrics;

	int fh;
	
	
	width_of_M= styles[st_def].font->getCharWidth('M');
	width_of_SPC= styles[st_code].font->getCharWidth('M');
	fh= styles[st_def].font->getFontHeight();

	m->page_width= getWidth()-verticalScrollBar()->getWidth();
	m->font_metrics= font_metrics;
	m->bullet_indent= bullet_indent;
	m->text_style= text_style;
	m->header_style= header_style;
	m->bullet_chr= 0x2022;
	m->paragraph_spacing= fh;
	m->link_style= docstyle_to_vstyle('L');
	m->preformatted_style= docstyle_to_vstyle('F');
	m->tabstop= tabstop;
	m->line_spacing= styles[st_def].font->getFontSpacing()/4; }


	if (parts)
		destroy_parts(nparts,parts);
	nparts= layout_doc(buffer, &metrics, &parts);
	 {
	int i;
	int maxx=0, maxy= 0;
	int minx= parts[0].x;
	int miny= parts[0].y;
	for(i=0;i<nparts;i++) {
		max2first(&maxx, parts[i].x+parts[i].width);
		max2first(&maxy,  parts[i].y+parts[i].height);
		min2first(&minx, parts[i].x);
		min2first(&miny, parts[i].y);
	}
	cont_width= maxx-minx;
	cont_height= maxy-miny;
	cont_x= minx;
	cont_y= miny;
}


}




 static  void font_metrics(layout_metrics_t *m,int style,wchar_t chr,
		int *fascent, int *fdescent, int *width) {
	*fascent= styles[style].font->getFontAscent();
	*fdescent= styles[style].font->getFontDescent();
	*width= styles[style].font->getCharWidth(chr);
}

 static  int bullet_indent(layout_metrics_t *m,int level) {
	return level * width_of_M * 4;
}

 static  int text_style(layout_metrics_t *m,int kind) {
	return docstyle_to_vstyle(kind); }

 static  int header_style(layout_metrics_t *m,int level) {
	char buf[20];
	sprintf(buf,"%d",level);
	return docstyle_to_vstyle(buf[0]);
}

 static  int tabstop(layout_metrics_t *m,int x) {
	int tsz= width_of_SPC*8;
	int i;
	for(i=0;i*tsz<=x;i++)
		;
	return i*tsz;
}



long xtview_t::onPaint(FXObject* ob,FXSelector sel,void* ptr){
	FXEvent* event=(FXEvent*)ptr;
	FXDCWindow dc(this,event);

	if (!nparts) {
		FXColor white;
		white= FXRGB(0xff,0xff,0xff);
		dc.setForeground(white);
		dc.fillRectangle(event->rect.x, event->rect.y,
			event->rect.w, event->rect.h);
		return 1;
	}

	int evX,evY;
	evX= event->rect.x - pos_x;
	evY= event->rect.y - pos_y;
	int evW= event->rect.w;
	int evH= event->rect.h;
	/* ok now evX,evY is the document coordinate */


	int s,e;
	int cx,cy,cw,ch;

	 {
	s= find_element(evX,evY, parts, nparts);
	e= find_element(evX+evW-1, evY+evH-1, parts,nparts);
	if (s==-1)
		s= 0;
	if (e==-1)
		e= nparts-1;
}


	 {
	int i;
	int minx= parts[s].x;
	int miny= parts[s].y;
	int maxx= parts[s].x+parts[s].width;
	int maxy= parts[s].y+parts[s].height;
	for(i=s+1;i<=e;i++) {
		min2first(&minx, parts[i].x);
		min2first(&miny, parts[i].y);
		max2first(&maxx, parts[i].x+parts[i].width);
		max2first(&maxy, parts[i].y+parts[i].height);
	}

	cx= minx;
	cy= miny;
	cw= maxx-minx;
	ch= maxy-miny;
}



	dc.setForeground(FXRGB(0xff,0xff,0xff));
	dc.fillRectangle(cx +pos_x, cy+pos_y, cw, ch);

	int cs= -1;
	int i;
	for(i=s;i<=e;i++) {
		if (parts[i].opcode=='T') {
			if (parts[i].d.text.style!=cs) {
				cs= parts[i].d.text.style;
				dc.setForeground(styles[cs].fgcolor);
				dc.setFont(styles[cs].font);
			}
			dc.drawText(parts[i].x+pos_x,
				parts[i].y +pos_y + parts[i].d.text.ascent,
				parts[i].d.text.text, strlen(parts[i].d.text.text));
		}
	}

	 {
	int ev_area[4];
	int painted_area[4];
	int dirty[16];
	int i;
	int n;
	i= 0;
	ev_area[i++]= evX; ev_area[i++]= evY;
	ev_area[i++]= evW; ev_area[i++]= evH;
	i= 0;
	painted_area[i++]= cx; painted_area[i++]= cy;
	painted_area[i++]= cw; painted_area[i++]= ch;
	
	n= rect_subtract(ev_area, painted_area, dirty);
	if (n)
		dc.setForeground(FXRGB(0xff,0xff,0xff));
	for(i=0;i<n;i++)
		dc.fillRectangle(pos_x+dirty[4*i+0], pos_y+dirty[4*i+1],
			dirty[4*i+2], dirty[4*i+3]);
/*
	printf("event area= %d %d %d %d\n",evX, evY, evW, evH);
	printf("content area= %d %d %d %d\n", cx,cy, cw,ch);

	for(i=0;i<n;i++)
		printf("%d %d %d %d\n", 
			dirty[4*i+0], 
			dirty[4*i+1], 
			dirty[4*i+2], 
			dirty[4*i+3]);
*/			
}




	return 1;
}



 static  int find_element(int x,int y,text_part_t *parts, int nparts) {
	int i;
	for(i=0;i<nparts;i++) 
		if (x>=parts[i].x && y>=parts[i].y &&
		    x<parts[i].x+parts[i].width &&
		    y<parts[i].y+parts[i].height)
		    	return i;
	return -1;
}

 




xtview_t::~xtview_t() {
	if (buffer)
		free(buffer);
	if (parts)
		destroy_parts(nparts,parts);
}

 static  void destroy_parts(int N,text_part_t *P) {
	int i;
	for(i=0;i<N;i++)
		if (P[i].opcode=='T')
			free(P[i].d.text.text);
	free(P);
}

void xtview_t::layout() {
	if (buffer)
		layout_buffer();
	FXScrollArea::layout();
}


 static  void min2first(int *A,int B) { if (B<*A) *A= B; }
 static  void max2first(int *A,int B) { if (B>*A) *A= B; }



 int rect_subtract(int *a,int *b,int *out) {
	int n=0;
	int aX= a[0], bX= b[0], aW= a[2], bW= b[2];
	int aY= a[1], bY= b[1], aH= a[3], bH= b[3];
	 
	out[ n++] = aX;
	out[ n++] = aY;
	out[ n++] = aW;
	out[ n++] = (bY-aY);
	if (!rect_sub_inside(a, out + n-4)) 
		n-= 4;

	 
	out[ n++] = aX;
	out[ n++] = (bY+bH);
	out[ n++] = aW;
	out[ n++] = (aX+aH-bY-bH);
	if (!rect_sub_inside(a, out + n-4)) 
		n-= 4;

	 
	out[ n++] = aX;
	out[ n++] = bY;
	out[ n++] = (bX-aX);
	out[ n++] = bH;
	if (!rect_sub_inside(a, out + n-4)) 
		n-= 4;

	 
	out[ n++] = (bX+bW);
	out[ n++] = bY;
	out[ n++] = (aX+aW-bX-bW);
	out[ n++] = bH;
	if (!rect_sub_inside(a, out + n-4)) 
		n-= 4;

	return n/4;
}




 static  int rect_sub_inside(int *a, int *o) {
	if (o[0] < a[0])  return 0;
	if (o[0] >= a[0]+a[2]) return 0;

	if (o[1] < a[1]) return 0;
	if (o[1] >= a[1]+a[3]) return 0;

	if (o[2]<=0 || o[3]<=0) return 0;
	return 1;
}
