
#include <stdlib.h>
#include <wctype.h>
#include <string.h>

#include "parser.h"
#include "utf8.h"

#include "layout.h"
 typedef struct {
	unsigned int glyph;
	unsigned short  x;
	unsigned int aux;
} xchar_t; typedef struct outbuf {
	struct outbuf *prev,*next;
	int nparts;
	text_part_t parts[40];
} outbuf_t; typedef struct {
	parser_sink_t vtable;
	layout_metrics_t *metrics;

	int nchars,schars;
	xchar_t *chars;

	int right;
	int left;
	int ascent, descent;
	int x,y;

	outbuf_t outbuf;

	int last_link_id;
} layout_t; typedef struct {
	wchar_t chr;
	unsigned short style;
	unsigned fascent,fdescent, width;
	unsigned char utf8[7];
} glyph_t; typedef struct glentry glentry_t;
struct glentry {
	unsigned index;
	int style;
	wchar_t chr;
	
  glentry_t *left, *right;
  int height;

};
 typedef struct {
	glentry_t *root;
} glindex_t;
 static  void fill_vtable(layout_t *L) ; static  void ignore_tag(layout_t *L, char* tag) ; static  void bullet(layout_t *L, int level) ; static  void pbreak(layout_t *L) ; static  void link(layout_t *L,int kind,char *txt,int len) ; static  void layout_text
	( layout_t *L,int style,char *txt, int len,int aux)
; static  void recalculate_row_dimensions(layout_t *L) ; static  void text(layout_t *L,int style,char *txt, int len) ; static  void formatted(layout_t *L,char *txt, int len) ; static  void header(layout_t *L,int level,char *txt, int len) ; static  void add_char(layout_t *L, unsigned g, int aux, int x) ; static  void max2first(int *A,int B) ; static  void finish_row(layout_t *L) ; static  void output_text_part(layout_t *L,int s,int e) ; static  int text_part_chars(xchar_t *d,int l,char *buf) ; static  text_part_t *collect_buffer_destroy(layout_t *L,int *Lr) ; static  int collect_buffer_nodestroy(outbuf_t *L,text_part_t *B) ; static  void
glindex_init(glindex_t *t) ; static  void*
glindex_find(glindex_t *t,glentry_t *key) ; static  int
glindex_insert(glindex_t *t,glentry_t *node) ; static  glentry_t*
glindex_insert_node(glentry_t *t,glentry_t *n) ; static  glentry_t*
glindex_balance(glentry_t *t) ; static  glentry_t*
glindex_remove(glindex_t *t, glentry_t *key) ; static  glentry_t*
glindex_remove_node(glentry_t *t,glentry_t *node) ; static  glentry_t*
glindex_rotate_right(glentry_t *g) ; static  glentry_t*
glindex_rotate_left(glentry_t *g) ; static  glentry_t*
glindex_remove_max(glentry_t *t,glentry_t **pl) ; static  glentry_t*
glindex_remove_min(glentry_t *t,glentry_t **pl) ; static  void
glindex_calc_height(glentry_t *t) ; static  glentry_t*
glindex_destroy(glindex_t *t) ; static  glentry_t*
glindex_destroy_one_node(glentry_t *node,glentry_t** plink) ; static  unsigned find_glyph(layout_t *L,wchar_t chr,int style) ; static  unsigned load_glyph(layout_metrics_t *metrics,wchar_t chr,int style) ; static  int glentry_compare(glentry_t *A,glentry_t *B) ;

	/* Although we dont have an allocation for glyph #0
	at the beginning, we are sure that it will not be
	accessed anyway. So setting nglyphs to 1 is safe.
	
	See also the grow array macro. If sglyphs is zero
	we dont check nglyphs so everything works. */
	static int nglyphs=1;
 static  int sglyphs; static  glyph_t *glyphs; static  glindex_t glindex;












 int layout_doc(char *txt,layout_metrics_t *metrics,text_part_t **result) {
	text_part_t *parts;
	int len;
	layout_t L;
	fill_vtable(&L);
	L.metrics= metrics;

	L.outbuf.prev= L.outbuf.next= &L.outbuf;
	L.outbuf.nparts= 0;

	L.nchars= 0;
	L.schars= 200;
	 
	L.chars= (typeof(L.chars)) calloc(L.schars, sizeof(L.chars[0]));


	L.x= 0;
	L.y= 0;

	L.ascent= L.descent= 0;

	L.left= 0;
	L.right= metrics->page_width;
	L.last_link_id= 1;

	parse_doc(txt, &L.vtable);
	finish_row(&L);

	parts= collect_buffer_destroy(&L,&len);
	free(L.chars);
	*result= parts;
	return len;
}

 static  void fill_vtable(layout_t *L) {
	parser_sink_t *s= &L->vtable;
	s->bullet= (void*) bullet;
	s->pbreak= (void*) pbreak;
	s->link= (void*) link;
	s->text= (void*) text;
	s->formatted= (void*) formatted;
	s->header= (void*) header;
	s->tag= (void*) ignore_tag;
}

 static  void ignore_tag(layout_t *L, char* tag) { }


 static  void bullet(layout_t *L, int level) {
	unsigned g;
	finish_row(L);
	L->left= L->metrics->bullet_indent(L->metrics, level);
	g= find_glyph(L, L->metrics->bullet_chr, 0);
	add_char(L, g, 0, L->left - glyphs[g].width );
	L->x= L->left;
}

 static  void pbreak(layout_t *L) {
	finish_row(L);
	L->left= 0;
	L->x= 0;
	L->y+= L->metrics->paragraph_spacing;
}

 static  void link(layout_t *L,int kind,char *txt,int len) {
	int style,aux;
	if (kind!='L')
		return ;
	/* FIXME: maybe use different styles for links
	   bound to Xs and Ws? */
	style= L->metrics->link_style;
	aux= L->last_link_id++;
	layout_text(L, style, txt,len, aux);
}


/* style here is translated style. i.e. device style */
 static  void layout_text
	( layout_t *L,int style,char *txt, int len,int aux)
{
	wchar_t *wstr;
	int i;
	wstr= utf8buf2wstr(txt,len);
	for(i=0;wstr[i];i++) {
		wchar_t chr;
		unsigned g;
		chr= wstr[i];
		if (!iswprint(chr))
			chr=' ';
		if (chr==' ' && L->x==L->left)
			continue;
		g= find_glyph(L,chr, style);
		if (L->x+glyphs[g].width>L->right)  {
			int j;
			int npurged;
			int first_purged;

			if (!L->nchars || L->x==L->left)
				goto add_anyway;
	
			if (iswspace(chr)) {
				/* trying to append a space at the end of a line
				and it doesn`t fit. so we ignore it :) */
				finish_row(L);
				continue;
			}

			if (iswspace(glyphs[L->chars[L->nchars-1].glyph].chr)) {
				/* the previous char was a space so we finish
				the row at where it is and then add the char */
				finish_row(L);
				goto add_anyway;
			}

			/* ok. at this point we know that the previous char was not
			a space and this one is not either. so, we got some back-
			tracking to do. */
			for(j=L->nchars-1;j>0;j--) 
				if (iswspace(glyphs[L->chars[j].glyph].chr))
					break;

			if (j==0) { 
				/* we found ourselves a great big word that doesnt
				fit anywhere, so we will char-wrap it instead. */
				finish_row(L);
				goto add_anyway;
			}
			first_purged= j+1;
			npurged= L->nchars-first_purged;

			/* ok. if this string is not long enough. i.e. different text()
			calls resulted in one continuous word to form, then backing
			up on the wstr array wont do any good because we can not
			get back to the original characters. so, I have to store the
			stuff in an auxilliary buffer and then re-transfer them to the
			row array. instead of using an auxilliary buffer, I will re-use
			the row array because I know that finish_row wont deallocate
			or overwrite those positions and I am a jerk. */
			L->nchars -= npurged;
			recalculate_row_dimensions(L);
			finish_row(L);
			
			for(j=0;j<npurged;j++) {
				xchar_t *xc= L->chars + j + first_purged;
				add_char(L, xc->glyph, xc->aux, L->x);
				L->x+= glyphs[xc->glyph].width;
			}
			/* now I silently fall thru to add the current character.
			    you cool with that? */
		}
	add_anyway:
		add_char( L, g, aux, L->x );
		L->x += glyphs[g].width;
	}
	free(wstr);
}

 static  void recalculate_row_dimensions(layout_t *L) {
	int i;
	L->ascent= L->descent= 0;
	for(i=0;i<L->nchars;i++) {
		max2first(&L->ascent, glyphs[L->chars[i].glyph].fascent);
		max2first(&L->descent, glyphs[L->chars[i].glyph].fdescent);
	}
}

 static  void text(layout_t *L,int style,char *txt, int len) {
	style= L->metrics->text_style(L->metrics,style);
	layout_text(L,style, txt,len, 0);
}

 static  void formatted(layout_t *L,char *txt, int len) {
	wchar_t *wstr;
	int i;
	int style;

	finish_row(L);
	L->left= 0;
	L->x= 0;
	style= L->metrics->preformatted_style;

	wstr= utf8buf2wstr(txt,len);
	for(i=0;wstr[i];i++) {
		wchar_t chr;
		int width;
		unsigned g;
		chr= wstr[i];
		if (chr=='\t') {
			int ts;
			g= find_glyph(L, ' ', style);
			ts= L->metrics->tabstop(L->metrics,L->x);
			width= ts-L->x;
		} else {
			if (chr=='\n')
				continue;
			g= find_glyph(L,chr,style);
			width= glyphs[g].width;
		}
		add_char(L, g, 0, L->x);
		L->x+= width;
	}
	free(wstr);
	finish_row(L);
}

 static  void header(layout_t *L,int level,char *txt, int len) {
	int style;
	finish_row(L);
	style= L->metrics->header_style(L->metrics, level);
	layout_text(L,style, txt,len,0);
	finish_row(L);
}




	








 static  void add_char(layout_t *L, unsigned g, int aux, int x) {
	xchar_t *xc;
	 {
	if (L->nchars==L->schars) {
		L->schars *= 2;
		
	L->chars= (typeof(L->chars)) realloc( L->chars, sizeof(L->chars[0])*(L->schars));


	}
}



	xc= L->chars+L->nchars;
	L->nchars++;
	xc->glyph= g;
	xc->aux= aux;
	xc->x= x;
	max2first(&L->ascent, glyphs[g].fascent);
	max2first(&L->descent, glyphs[g].fdescent);
}

 static  void max2first(int *A,int B) { if (B>*A) *A= B; }

 static  void finish_row(layout_t *L) {
	int s,e;
	if (L->nchars) {
		s= 0;

	again:
		e= s+1;
		while(e<L->nchars) {
			if (glyphs[L->chars[e].glyph].style!=glyphs[L->chars[s].glyph].style)
				break;
			if (L->chars[e].x!=L->chars[e-1].x+glyphs[L->chars[e-1].glyph].width)
				break;
			if (L->chars[e].aux!=L->chars[s].aux)
				break;
			e++;
		}

		output_text_part(L,s,e);
		if (e<L->nchars) {
			s= e;
			goto again;
		}
	}

	L->nchars= 0;
	L->x= L->left;
	L->y= L->y + L->ascent + L->descent + L->metrics->line_spacing;
	L->ascent= L->descent= 0;
}

 static  void output_text_part(layout_t *L,int s,int e) {
	text_part_t  T;
	int len;
	outbuf_t *O;
	T.opcode= 'T';
	T.y= L->y;
	T.d.text.ascent= L->ascent;
	T.d.text.aux=L->chars[s].aux;
	T.d.text.style= glyphs[L->chars[s].glyph].style;
	T.x= L->chars[s].x;
	len= text_part_chars(L->chars+s, e-s, 0);
	T.d.text.text= malloc(len);
	text_part_chars(L->chars+s, e-s, T.d.text.text);

	T.height= L->ascent + L->descent;
	T.width= L->chars[e-1].x
		+ glyphs[L->chars[e-1].glyph].width
		- L->chars[s].x;

	O= L->outbuf.prev;
	if (O->nparts==sizeof(O->parts)/sizeof(T)) {
		
	O= (typeof(O)) calloc(1, sizeof(O[0]));


		 {
	(O)->prev= ((&L->outbuf))->prev; 
	(O)->next= ((&L->outbuf)); 
	((&L->outbuf))->prev->next= (O); 
	((&L->outbuf))->prev= (O); 
}


	}
	O->parts[O->nparts++]= T;
}


 static  int text_part_chars(xchar_t *d,int l,char *buf) {
	int i;
	int L=0;
	for(i=0;i<l;i++) {
		char *u= (char*) glyphs[d[i].glyph].utf8;
		int xl= strlen(u);
		if (buf)
			memcpy(buf+L, u, xl);
		L+= xl;
	}
	if (buf)
		buf[L]= 0;
	return L+1;
}









 static  text_part_t *collect_buffer_destroy(layout_t *L,int *Lr) {
	int len;
	text_part_t *P;
	len= collect_buffer_nodestroy(&L->outbuf, 0);
	 
	P= (typeof(P)) calloc(len, sizeof(P[0]));


	collect_buffer_nodestroy(&L->outbuf, P);
	while(1) {
		outbuf_t *B= L->outbuf.prev;
		if (B==&L->outbuf)
			break;
		 {
	(B)->prev->next= (B)->next; 
	(B)->next->prev= (B)->prev; 
}


		free(B);
	}
	*Lr= len;
	return P;
}


 static  int collect_buffer_nodestroy(outbuf_t *L,text_part_t *B) {
	outbuf_t *N= L;
	int s= 0;
	do {
		if (B && N->nparts) 
			memcpy(B+s, N->parts,
				N->nparts*sizeof(N->parts[0]));
		s+= N->nparts;
		N= N->next;
	} while(N!=L);
	return s;
}










 static  void
glindex_init(glindex_t *t) { t->root= 0; }


 static  void*
glindex_find(glindex_t *t,glentry_t *key) {
	glentry_t *n= t->root;
	while(n) {
		int u;
		u= glentry_compare(key,n);
		if (u==0) return n;
		else if (u<0) n=n->left;
		else n=n->right;
	}
	return 0;
}

 static  int
glindex_insert(glindex_t *t,glentry_t *node) {
	if (glindex_find(t,node)) {
		return 1;
	} else {
		node->height= 1;
		node->left= node->right= 0;
		t->root= glindex_insert_node(t->root,node);
		return 0;
	}
}

 static  glentry_t*
glindex_insert_node(glentry_t *t,glentry_t *n) {
	int u;
	if (!t)
		return n;
	u= glentry_compare(n,t);
	if (u<0)
		t->left= glindex_insert_node(t->left,n);
	else
		t->right= glindex_insert_node(t->right,n);
	return glindex_balance(t);
}

#ifndef TREE_height
#define TREE_height(x) ((x)? (x)->height : 0)
#endif

 static  glentry_t*
glindex_balance(glentry_t *t) {
	int u;
	glindex_calc_height(t);
	u= TREE_height(t->left)-TREE_height(t->right);
	if (u>1)
		return glindex_rotate_right(t);
	else if (u<-1)
		return glindex_rotate_left(t);
	else return t;
}

 static  glentry_t*
glindex_remove(glindex_t *t, glentry_t *key) {
	glentry_t *node;
	node= glindex_find(t,key);
	if (node)
		t->root= glindex_remove_node(t->root,node);
	return node;
}

 static  glentry_t*
glindex_remove_node(glentry_t *t,glentry_t *node) {
	int u;
	u= glentry_compare(node,t);
	if (u<0) 
		t->left= glindex_remove_node(t->left,node);
	else if (u>0)
		t->right= glindex_remove_node(t->right,node);
	else {
		if (!t->right || !t->left) {
			if (t->right)
				return t->right;
			else
				return t->left;
		} else {
			glentry_t *p;
			if (t->left->height>t->right->height)
				p= glindex_remove_max(t->left, &t->left);
			else 
				p= glindex_remove_min(t->right, &t->right);
			p->left= t->left;
			p->right= t->right;
			p->height=1;
			t= p;
		}
	}
	return glindex_balance(t);
}

 static  glentry_t*
glindex_rotate_right(glentry_t *g) {
	glentry_t *f,*e;
	f= g->left;
	e= f->right;
	f->right= g;
	g->left= e;
	glindex_calc_height(g);
	glindex_calc_height(f);
	return f;
}

 static  glentry_t*
glindex_rotate_left(glentry_t *g) {
	glentry_t *f,*e;
	f= g->right;
	e= f->left;
	f->left= g;
	g->right= e;
	glindex_calc_height(g);
	glindex_calc_height(f);
	return f;
}


 static  glentry_t*
glindex_remove_max(glentry_t *t,glentry_t **pl) {
	if (t->right) {
		glentry_t *z;
		z= glindex_remove_max(t->right,&t->right);
		glindex_calc_height(t);
		return z;
	}
	*pl= t->left;
	return t;
}

 static  glentry_t*
glindex_remove_min(glentry_t *t,glentry_t **pl) {
	if (t->left) {
		glentry_t *z;
		z= glindex_remove_min(t->left,&t->left);
		glindex_calc_height(t);
		return z;
	}
	*pl= t->right;
	return t;
}

 static  void
glindex_calc_height(glentry_t *t) {
	int a,b;
	a= TREE_height(t->left)+1;
	b= TREE_height(t->right)+1;
	if (a>b) t->height=a;
	else t->height= b;
}

 static  glentry_t*
glindex_destroy(glindex_t *t) {
	if (!t->root)
		return 0;
	return glindex_destroy_one_node(t->root,&t->root);
}

 static  glentry_t*
glindex_destroy_one_node(glentry_t *node,glentry_t** plink) {
	if (node->left)
		return glindex_destroy_one_node(node->left, &node->left);
	else if (node->right)
		return glindex_destroy_one_node(node->right, &node->right);
	*plink= 0;
	return node;
}












 static  unsigned find_glyph(layout_t *L,wchar_t chr,int style) {
	glentry_t *ge;
	glentry_t key;
	key.style= style;
	key.chr= chr;
	ge= glindex_find(&glindex, &key);
	if (ge)
		return ge->index;
	
	ge= (typeof(ge)) calloc(1, sizeof(ge[0]));


	ge->style= style;
	ge->chr= chr;
	ge->index= load_glyph(L->metrics, chr, style);
	glindex_insert(&glindex, ge);
	return ge->index;
}

 static  unsigned load_glyph(layout_metrics_t *metrics,wchar_t chr,int style) {
	glyph_t *G;
	 {
	if (!sglyphs) {
		sglyphs= 256;
		 
	glyphs= (typeof(glyphs)) calloc(sglyphs, sizeof(glyphs[0]));


	} else if (nglyphs==sglyphs) {
		sglyphs += sglyphs/2;
		
	glyphs= (typeof(glyphs)) realloc( glyphs, sizeof(glyphs[0])*(sglyphs));


	}
}


	G= glyphs+nglyphs;
	G->chr= chr;
	G->style= style;
	metrics->font_metrics(metrics, style, chr, 
		(int*) &G->fascent, (int*) &G->fdescent, (int*) &G->width);
	wchr2utf8str(chr, (char*) G->utf8);
	return nglyphs++;
}




 static  int glentry_compare(glentry_t *A,glentry_t *B) {
	if (A->style==B->style)
		return A->chr - B->chr;
	return A->style-B->style;
}

void layout_glindex_unused_functions() {
	glindex_destroy(0);
	glindex_init(0);
	glindex_remove(0,0);
}


