
#include <ctype.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>



 typedef struct parser_sink parser_sink_t;
 void parse_doc(char *str,parser_sink_t *sink) ; void
file_read_all(char *fn,char **buf,int *len) ;

 typedef struct {
	char *str;
	parser_sink_t *sink;
} parser_t;

 static  void output_bullet_pbreak(parser_t *p,int d) ; static  void output_paragraph_break(parser_t *p) ; static  void output_header(parser_t *p,int level, char *s,char *e) ; static  void output_link(parser_t *p,int kind,char *s,char *e) ; static  void output_text(parser_t *p,int kind, char *s,char *e) ; static  void output_tag(parser_t *p,char *value)
; static  void
prt_range(char *s,char *e) ; static  void
output_preformatted(parser_t *p,char *s,char *e) ; static  void parse_bullet(parser_t *p) ; static  void paragraph_break(parser_t *p) ; static  void parse_header(parser_t *p) ; static  void parse_preformatted(parser_t *p) ; static  void parse_text(parser_t *p) ; static  void parse_inline_attribs(parser_t *p) ; static  void parse_normal_text(parser_t *p) ; static  void parse_attrib(parser_t *p) ; static  char **
strsplit(char *re,char *d) ; static  char *
collect_substr(char *str,int len) ; static  void do_attrib_line(parser_t *p,char *line)
;





 static  void output_bullet_pbreak(parser_t *p,int d) { printf("BP[%d]\n",d); }
 static  void output_paragraph_break(parser_t *p) {printf("P\n"); }
 static  void output_header(parser_t *p,int level, char *s,char *e) {
	printf("HDR[%d]<",level);
	prt_range(s,e);
	printf(">\n");
}
 static  void output_link(parser_t *p,int kind,char *s,char *e) {
	printf("LNK[%d]<",kind);
	prt_range(s,e);
	printf(">\n");
}

 static  void output_text(parser_t *p,int kind, char *s,char *e) {
	if (!kind) kind= 'D';
	printf("TXT[%c]<",kind);
	prt_range(s,e);
	printf(">\n");
}

 static  void output_tag(parser_t *p,char *value)
{
	printf("TAG[%s]\n",value);
}

int main(int argc,char **argv) {
	char *buf;
	int len;
	file_read_all(argv[1],&buf,&len);
	parse_doc(buf,0);
	return 0;
}

 static  void
prt_range(char *s,char *e) {
	fwrite(s, 1, e-s, stdout);
}

 static  void
output_preformatted(parser_t *p,char *s,char *e) {
	printf("PRE\n");
	prt_range(s,e);
	printf("END\n");
}




 void parse_doc(char *str,parser_sink_t *sink) {
	parser_t parser;
	parser_t *p= &parser;
	int own_buffer;
	parser.sink= sink;

	 {
	int L;
	int i;
	own_buffer= 0;
	L= strlen(str);
	for(i=0;i<L;i++)
		if (str[i]=='\r') {
			own_buffer=1;
			break;
		}
	if (own_buffer) {
		int j;
		char *nb= malloc(L+1);
		j= 0;
		for(i=0;i<=L;i++)
			if (str[i]!='\r') nb[j++]= str[i];
		str= nb;
	}
}
			



	parser.str= str;

	while(parser.str[0]) {
		switch(parser.str[0]) {
			case '\n': paragraph_break(p); break;
			case '=': parse_header(p); break;
			case '+': parse_bullet(p); break;
			case ' ': parse_preformatted(p); break;
			case '\t': parse_preformatted(p); break;
			case '@': parse_attrib(p); break;
			default: parse_text(p); 
		}
	}
	if (own_buffer)
		free(str);
}

 static  void parse_bullet(parser_t *p) {
	int level=0;
	while(p->str[0]=='+') {
		level++;
		p->str++;
	}
	if (p->str[0]=='\n')
		return ;

	while(1) {
		switch(p->str[0]) {
		case '\n': p->str++;
		case 0: return ;
		default:
			if (isspace(p->str[0])) 
				p->str++;
			else {
				output_bullet_pbreak(p,level);
				parse_text(p);
				return ;
			}
		}
	}
}


 static  void paragraph_break(parser_t *p) {
	output_paragraph_break(p);
	p->str++;
}


 static  void parse_header(parser_t *p) {
	char *s, *e;
	int level=0;
	int state= 'I';
	while(state!='X') {
		switch(state) {
		case 'I': switch(p->str[0]) {
			case '=': level++; p->str++; break;
			case '\n': p->str++; return ;
			case 0: return ;
			default: state= 'C'; s= p->str; p->str++;
			} break;
		case 'C': switch(p->str[0]) {
			case '=': e= p->str; state= 'S'; break;
			case '\n': e= p->str; state= 'N'; break;
			case 0: e= p->str; state= 'X'; break;
			default: p->str++;
			} break;
		case 'S': switch(p->str[0]) {
			case '=': p->str++; break;
			case 0: state= 'X'; break;
			case '\n': state= 'N'; break;
			default: p->str++;
			} break;
		case 'N': p->str++; state= 'X'; break;
		}
	}
	if (level>4) level= 4;
	output_header(p, 5-level, s,e);
}

 static  void parse_preformatted(parser_t *p) {
	char *s,*e;
	int done= 0;
	s= p->str;
	do {
		switch(p->str[0]) {
		case '\n': p->str++; e= p->str; done= 1; break;
		case 0: e= p->str; done= 1; break;
		default: p->str++;
		}
	} while(!done);
	output_preformatted(p, s,e);
}

 static  void parse_text(parser_t *p) {
	int done= 0;
	do {
		switch(p->str[0]) {
		
	case 'L': case 'X': case 'W':


		
	case 'R': case 'B': case 'I':
	case 'G': case 'C':


			parse_inline_attribs(p);
			break;
		case '\n':
			p->str++;
		case 0:
			done= 1;
			break;
		default: 
			parse_normal_text(p);
		}
	} while(!done);
}

 static  void parse_inline_attribs(parser_t *p) {
	int c;
	int b;
	int done;
	char *s,*e;
	c= p->str[0];
	p->str++;
	if (p->str[0]!='{' && p->str[0]!='<') {
		output_text(p, 'D', p->str-1, p->str);
		return ;
	}
	b= p->str[0];
	p->str++;
	s= p->str;
	done= 0;
	do {
		switch(p->str[0]) {
		case 0:
			e= p->str;
			done= 1;
			break;
		case '>': 
			if (b=='<') {
				e= p->str;
				p->str++;
				done= 1;
			} else p->str++;
			break;
		case '}':
			if (b=='{') {
				e= p->str;
				p->str++;
				done= 1;
				break;
			}
		default:
			p->str++;
		}
	} while(!done);
	switch(c) {
	
	case 'L': case 'X': case 'W':


		output_link(p,c,s,e);
		break;
	
	case 'R': case 'B': case 'I':
	case 'G': case 'C':


		output_text(p, c,s,e);
		break;
	}
}

/* this function doesnt eat the newline so that 
the parent can look at it. */
 static  void parse_normal_text(parser_t *p) {
	char *s,*e;
	int done= 0;
	s= p->str;
	do {
		switch(p->str[0]) {
		case '\n':
		case 0:
		
	case 'L': case 'X': case 'W':


		
	case 'R': case 'B': case 'I':
	case 'G': case 'C':


			e= p->str; 
			if (*e=='\n') e++;
			done= 1; 
			break;
		default:
			p->str++;
		}
	} while(!done);
	output_text(p, 'D', s,e);
}

/* ok. so, 
	\n\n -> paragraph break
	\n= -> header
	\n+ -> list
	\n\t -> preformatted
	\nSPC -> preformatted
	\n.  -> normal text

	ok. so just assume that we enter this function only
	at the beginning of a line. cool.
	
	if a piece of styled text spans multiple lines, then pass those
	lines within the styled text and  then come back here. ok.

	we assume that the newline has already been eaten.

	we need a little bit of buffering at the output so that it makes
	sense. ie. outputting two paragraph breaks in a line does not
	make sense. neither does a paragraph break followed by a
	heading. (the paragraph break should be removed, then)
*/








 static  void parse_attrib(parser_t *p) {
	char *s, *e;
	char *line;
	s= p->str;
	while(p->str[0]!='\n' && p->str[0]!=0)
		p->str++;
	e= p->str;
	if (e-s>1) {
		line= malloc(e-s+1);
		memcpy(line, s, e-s);
		line[e-s]= 0;
		do_attrib_line(p,line);
		free(line);
	}
}



 static  char **
strsplit(char *re,char *d) {
	regex_t rx;
	int ec;
	if ((ec=regcomp(&rx,re,REG_EXTENDED))) {
		char ebuf[2000];
		regerror(ec,&rx, ebuf, sizeof(ebuf)-1);
		exit(fprintf(stderr,
			"regcomp for '%s' failed: %s\n", re, ebuf));
	}
	char *orig_d= d;
	int nfields;
	char **fields=0;

again:
	d= orig_d;
	nfields= 0;
	while(*d) {
		regmatch_t matches[2];
		if (regexec(&rx,d, 2, matches, 0)) {
			if (fields)
				fields[nfields]= collect_substr(d,
						 strlen(d));
			nfields++;
			d+= strlen(d);
		} else {
			if (matches[0].rm_so==0) {
				d+= matches[0].rm_eo;
			} else {
				if (fields) 
					fields[nfields]= collect_substr(d,
							 matches[0].rm_so);
				d+= matches[0].rm_eo;
				nfields++;
			}
		}
	}
	if (fields) {
		regfree(&rx);
		return fields;
	}
	if (!nfields) {
		regfree(&rx);
		return 0;
	}
	 
	fields= (typeof(fields)) calloc((nfields+1), sizeof(fields[0]));


	fields[nfields]= 0;
	goto again;
}

 static  char *
collect_substr(char *str,int len) {
	char *x;
	 
	x= (typeof(x)) calloc((len+1), sizeof(x[0]));


	memcpy(x,str,len);
	x[len]= 0;
	return x;
}








 static  void do_attrib_line(parser_t *p,char *line)
{
	char **elts;
	int i;
	elts= strsplit("[ \t\v\b\r\f\r]+", line);
	if (!elts)
		return ;
	if (!elts[0] || strcmp(elts[0],"@tags"))
		goto done;
	for(i=1;elts[i];i++)
		output_tag(p, elts[i]);
done:
	for(i=0;elts[i];i++)
		free(elts[i]);
	free(elts);
}










 void
file_read_all(char *fn,char **buf,int *len) {
	size_t L;
	struct stat sb;
	FILE *f;
	if (stat(fn,&sb))
		exit(fprintf(stderr,"can not stat %s: %s\n",
			fn,strerror(errno)));
	f= fopen(fn,"rb");
	if (!f)
		exit(fprintf(stderr,"can not open %s: %s\n",
			fn,strerror(errno)));
	L= sb.st_size;
	*len= 0;
	*buf= (char*) malloc(L+1);
	while((size_t)*len<L) {
		size_t r= fread(*buf + *len, 1, L-*len, f);
		if (!r)
			exit(fprintf(stderr,"can not read %s: %s\n",
				fn,strerror(errno)));
		*len += r;
	}
	(*buf)[L]= 0;
	fclose(f);
}
