-[ c library "main.c", header "main.h" ]


int main(int argc,char **argv) {
	-[private] page_t *pageset;
	page_t *p;
	if (argc>1) 
		load_file_nodisplay(argv[1]);
	if (!pageset)
		pageset= make_default_doc();
	p= get_default_page(pageset);
	get_link_targets(p->text);
	init_display(p->name,p->text);
	set_node_name(p->name);
	display_tags(p->text);
	history_push(p->name);
	run_gui();
	return 0;
}

-[ lheader "display.h" ]
-[ lheader "docparse.h" ]
-[ header "string.h" ]
-[ header "ctype.h" ]
-[ header "stdlib.h" ]
-[ lheader "links.h" ]
-[ lheader "toc.h" ]
-[ lheader "parser.h" ]
-[ lheader "search.h" ]


+[private variables]
static char default_doc[]= 
"=== Main Page ==="
"\n\n"
"This is the main page of your document. "
"Edit this page to add your own content." 
;
+[ end ]
	

-[private] page_t *
make_default_doc() {
	page_t *root;
	page_t *p;
	-[ allocate object root ]
	root->prev= root->next= root;
	-[ allocate object p ]
	p->text= strdup(default_doc);
	p->name= strdup("main");
	-[ clist/insert p before root ]
	return root;
}

-[private] page_t* get_default_page(page_t *pages) {
	page_t *p;
	p= find_page("main");
	if (p)
		return p;
	else
		return pages->next;
}

-[public] void follow_link(int id) {
	char *tgt;
	tgt= link_target(id);
	if (!tgt)
		return ;
	history_push(tgt);
	goto_page(tgt,link_text(id));
}

-[private] void goto_page(char *tgt,char *xtitle) {
	page_t *p;
	p= find_page(tgt);
	reset_tag_display();
	if (p) {
		display_text(tgt,p->text);
		set_node_name(p->name);
		get_link_targets(p->text);
		display_tags(p->text);
	} else {
		char buf[2000];
		if (xtitle) {
			sprintf(buf,"=== %s ===\n"
				"The page %s does not exist in the "
				"document.\n\n"
				"In order to create it, edit this page "
				"and then save.", xtitle, tgt);
		} else {
			sprintf(buf,"The page %s does not exist in the "
				"document.\n\n"
				"In order to create it, edit this page "
				"and then save.", tgt);
		}
		display_text(tgt,buf);
		set_node_name(tgt);
		get_link_targets(buf);
	}
}

-[private] void print_tag(void *parser,char *text) 
{
	add_page_tag(text);
}

-[private] void ignore_bullet(void*a1,int a2) { }
-[private] void ignore_pbreak(void*a1) { }
-[private] void ignore_text(void *a1,int a2,char *a3,int a4) { }
-[private] void ignore_formatted(void *a1,char *a2,int a3) { }
-[private] void ignore_header(void *a1,int a2,char *a3,int a4) { }
-[private] void ignore_link(void *a1,int k,char* v,int l) { }

-[public] void display_tags(char *text) {
	parser_sink_t sink;
	sink.bullet= ignore_bullet;
	sink.pbreak= ignore_pbreak;
	sink.link= ignore_link;
	sink.text= ignore_text;
	sink.formatted= ignore_formatted;
	sink.header= ignore_header;
	sink.tag= print_tag;
	parse_doc(text, &sink);
}


-[private] page_t *find_page(char *name) {
	page_t *p;
	-[ for each p in list pageset, begin ]
		if (!strcmp(p->name,name))
			return p;
	-[ end ]
	return 0;
}

-[private] void set_node_name(char *name) {
	-[ private] char *node_name;
	if (node_name)
		free(node_name);
	node_name= strdup(name);
}

-[ header "stdio.h" ]

-[public] void save_page(char *name,char *text) {
	page_t *p;
	p= find_page(name);
	if (p) {
		free(p->text);
		p->text= strdup(text);
	} else {
		-[ allocate object p ]
		p->name= strdup(name);
		p->text= strdup(text);
		-[clist/insert p before pageset]
	}
	set_node_name(name);
	display_text(p->name, p->text);
	get_link_targets(p->text);
	reset_tag_display();
	display_tags(p->text);
}

-[private] void load_file_nodisplay(char *fname) {
	page_t *p;
	set_string(&file_name,fname);
	if (pageset) {
		-[ for each p in list pageset, begin ]
			free(p->name);
			free(p->text);
			-[ clist/remove p ]
			free(p);
		-[ end ]
		free(pageset);
	}
	pageset= read_doc(fname);
	if (!pageset)
		pageset= make_default_doc();
}

-[public] void load_file(char* fname) {
	page_t *p;
	-[private] char *file_name;
	load_file_nodisplay(fname);
	p= get_default_page(pageset);
	get_link_targets(p->text);
	display_text(p->name, p->text);
}

-[private] void set_string(char **str,char *val) {
	if (*str!=val) {
		if (*str)
			free(*str);
		*str= strdup(val);
	}
}

-[public] char* get_file_name() { return file_name ? strdup(file_name) : 0 ; }

-[public] void save_file(char *fname) {
	if (save_doc(fname,pageset))
		return ;
	set_string(&file_name,fname);
}

-[public] void move_back() {
	histent_t *p;
	p= history->prev;
	if (p==&history_store)
		return ;
	history= p;
	goto_page(p->name);
}

-[public] void move_forward() {
	histent_t *p;
	p= history->next;
	if (p==&history_store)
		return ;
	history= p;
	goto_page(p->name);
}

-[private] void history_push(char *name) {
	histent_t *p,*N;
	for(p=history->next;p!=&history_store;p=N) {
		N= p->next;
		-[clist/remove p]
		free(p->name);
		free(p);
	}
	-[ allocate object p ]
	p->name= strdup(name);
	-[clist/insert p before (&history_store) ]
	history= p;
}


-[private] typedef struct hent {
	struct hent *prev,*next;
	char *name;
} histent_t;

+[private variables]
static histent_t history_store= {&history_store, &history_store};
static histent_t *history = &history_store;
+[end]

-[ public] void display_toc() {
	char *text;
	text= gen_toc(pageset);
	display_text(":history", text);
	get_link_targets(text);
	free(text);
	reset_tag_display();
}


-[public] void do_tag_search(char* expr)
{
	char *text;
	int i;
	for(i=0;expr[i];i++)
		if (!isspace(expr[i]))
			break;
	if (!expr[i]) return ;
	text= search_tags(pageset,expr);
	display_text(":search", text);
	get_link_targets(text);
	free(text);
	reset_tag_display();
}
