
-[private] typedef struct {
	wchar_t chr;
	unsigned short style;
	unsigned fascent,fdescent, width;
	unsigned char utf8[7];
} glyph_t;

+[private variables]
	/* Although we dont have an allocation for glyph #0
	at the beginning, we are sure that it will not be
	accessed anyway. So setting nglyphs to 1 is safe.
	
	See also the grow array macro. If sglyphs is zero
	we dont check nglyphs so everything works. */
	static int nglyphs=1;
+[end]
-[private] int sglyphs;
-[private] glyph_t *glyphs;


-[ instantiate tree glindex_t :
	scope= private ,
	node type= glentry_t,
	comparision= glentry_compare,
	prefix= glindex ]


-[private] typedef struct glentry glentry_t;

+[private types]
struct glentry {
	unsigned index;
	int style;
	wchar_t chr;
	-[ tree members for glentry_t ]
};
+[ end ]

-[private] typedef struct {
	glentry_t *root;
} glindex_t;

-[private] glindex_t glindex;

-[private] unsigned find_glyph(layout_t *L,wchar_t chr,int style) {
	glentry_t *ge;
	glentry_t key;
	key.style= style;
	key.chr= chr;
	ge= glindex_find(&glindex, &key);
	if (ge)
		return ge->index;
	-[ allocate object ge ]
	ge->style= style;
	ge->chr= chr;
	ge->index= load_glyph(L->metrics, chr, style);
	glindex_insert(&glindex, ge);
	return ge->index;
}

-[private] unsigned load_glyph(layout_metrics_t *metrics,wchar_t chr,int style) {
	glyph_t *G;
	-[ make space in glyphs ]
	G= glyphs+nglyphs;
	G->chr= chr;
	G->style= style;
	metrics->font_metrics(metrics, style, chr, 
		(int*) &G->fascent, (int*) &G->fdescent, (int*) &G->width);
	wchr2utf8str(chr, (char*) G->utf8);
	return nglyphs++;
}


=[ make space in glyphs ] {
	if (!sglyphs) {
		sglyphs= 256;
		-[ glyphs = allocate array sglyphs ]
	} else if (nglyphs==sglyphs) {
		sglyphs += sglyphs/2;
		-[ reallocate array glyphs to size sglyphs ]
	}
}

=[]

-[private] int glentry_compare(glentry_t *A,glentry_t *B) {
	if (A->style==B->style)
		return A->chr - B->chr;
	return A->style-B->style;
}

void layout_glindex_unused_functions() {
	glindex_destroy(0);
	glindex_init(0);
	glindex_remove(0,0);
}


