-[ c library "links.c", header "links.h" ]


-[public] void get_link_targets(char *txt) {
	parser_sink_t sink;
	-[ reset links array ]
	-[ setup parser sink ]
	parse_doc(txt, &sink);
}

-[ private] typedef struct {
	char *target;
	char *text;
} link_t;

=[ reset links array ]
	-[private] int nlinks;
	-[private] int slinks;
	-[private] link_t *links;
	int i;
	for(i=0;i<nlinks;i++) {
		free(links[i].target);
		if (links[i].text)
			free(links[i].text);
	}
	nlinks= 0;

=[ setup parser sink ]
	sink.bullet= ignore_bullet;
	sink.pbreak= ignore_pbreak;
	sink.link= record_link;
	sink.text= ignore_text;
	sink.formatted= ignore_formatted;
	sink.header= ignore_header;
	sink.tag= ignore_tag;

=[]
-[private] void ignore_bullet(void*a1,int a2) { }
-[private] void ignore_pbreak(void*a1) { }
-[private] void ignore_text(void *a1,int a2,char *a3,int a4) { }
-[private] void ignore_formatted(void *a1,char *a2,int a3) { }
-[private] void ignore_header(void *a1,int a2,char *a3,int a4) { }
-[private] void ignore_tag(void *a1,char* v) { }

-[private] void record_link(void *arg,int kind,char *txt,int len) {
	switch(kind) {
	case 'L':
		if (!links[nlinks-1].text)
			links[nlinks-1].text= buf2str(txt,len);
		break;
	case 'X':
		-[ make space in links ]
		links[nlinks].target= buf2str(txt,len);
		links[nlinks].text= 0;
		nlinks++;
		break;
	}
}


=[ make space in links ]
	if (nlinks==slinks) {
		if (!slinks) {
			slinks= 50;
			-[ links= allocate array slinks ]
		} else {
			slinks*= 2;
			-[ reallocate array links to size slinks ]
		}
	}

=[]

-[private] char *buf2str(char *buf,int len) {
	char *s;
	s= malloc(len+1);
	memcpy(s, buf,len);
	s[len]= 0;
	return s;
}

-[ header "stdlib.h" ]
-[ header "string.h" ]

-[public] char* link_target(int id) {
	if (id<1 || id>nlinks)
		return 0;
	return links[id-1].target;
}

-[public] char* link_text(int id) {
	if (id<1 || id>nlinks)
		return 0;
	return links[id-1].text;
}

-[ lheader "parser.h" ]

