-[ c library "search.c", header "search.h" ]

-[ headers, begin ]
	!stdlib !string docparse parser
-[ end ]

-[ import strbuf library in private scope ]

-[private] typedef struct {
	page_t *page;
	char *title;
	char *tags;
} entry_t;

-[ selist_t : private list of entry_t* ]

-[public] char* search_tags(page_t *L,char *expr)
{
	page_t *p;
	selist_t *ser;
	strbuf_t *rbuf;

	ser= selist_new();
	-[ nlist>for each p in L, begin ]
		try_match(ser,p, expr);
	-[ end ]
	if (selist_isempty(ser)) {
		char buf[2000];
		free(ser);
		sprintf(buf,"=== Search Results ===\n\n"
			"The expression '%s' did not match any of the "
			"pages.\n", expr);
		return strdup(buf);
	}
	-[list>qsort ser:selist_t, begin]
		result= strcmp( E1->title, E2->title );
	-[end]
	rbuf= strbuf_init();
	sbprintf(rbuf,"=== Search Results ===\n\n");
	sbprintf(rbuf,"Search Expression= '%s'\n\n", expr);
	while(!selist_isempty(ser)) {
		entry_t *E= selist_destroy(ser);
		sbprintf(rbuf,"+ X<%s>L<%s> : %s\n",
			E->page->name, E->title, E->tags);
		free(E->title);
		free(E->tags);
		free(E);
	}
	sbprintf(rbuf,"+ \n");
	return strbuf_collect(rbuf);
}

-[private] void ignore_bullet(void*a1,int a2) { }
-[private] void ignore_pbreak(void*a1) { }
-[private] void ignore_text(void *a1,int a2,char *a3,int a4) { }
-[private] void ignore_formatted(void *a1,char *a2,int a3) { }
-[private] void ignore_link(void *a1,int k,char* v,int l) { }

-[private] void try_match(selist_t *ret,page_t *page,char *expr)
{
	entry_t *E;
	-[private] char query[200];
	-[private] char page_title[2000];
	-[private] int page_title_level;
	-[private] char page_tags[2000];
	parser_sink_t sink;
	-[ prepare the sink ]
	strcpy(query,expr);
	page_tags[0]= 0;
	page_title_level= 100;
	page_title[0]= 0;
	parse_doc(page->text, &sink);

	-[ evaluate clauses about unknown tags ]
	if (!eval_expr(query))
		return ;

	-[ allocate object E ]
	E->page= page;
	if (page_title[0])
		E->title= strdup(page_title);
	else
		E->title= strdup(page->name);
	if (page_tags[0])
		E->tags= strdup(page_tags);
	else
		E->tags= strdup("");
	selist_append(ret, E);
}

-[private] void check_header(void *a1,int level,char *str,int len)
{
	if (page_title_level<=level)
		return ;
	memcpy(page_title, str,len);
	page_title[len]= 0;
	page_title_level= level;
}

-[private] void check_tag(void *sink, char *tag)
{
	char t[200];
	strcat(page_tags,tag);
	strcat(page_tags, " ");

	sprintf(t,"!%s",tag);
	while(subs(query,t,"0"))   
		;
	sprintf(t,"=%s",tag);
	while(subs(query,t,"1"))
		;
}

-[private] int get_arg
	(char *expr,char *op, char **begin,char **end, char *res)
{
	int i;
	char *s;
	char *d;
	s= strstr(expr,op);
	if (!s)
		return 0;
	*begin= s;
	d= s+strlen(op);
	for(i=0;d[i];i++)
		if (d[i]==')' || d[i]=='&' || d[i]=='|') break;
	*end= d+i;
	for(s=d;s<*end;s++)
		*(res++)= *s;
	*res= 0;
	return 1;
}

-[private] void replace(char *begin,char *end, char *rep) {
	int len;
	len= strlen(rep);
	memcpy(begin,rep,len);
	for(begin=begin+len;begin<end;begin++)
		*begin=' ';
}

-[private] int compact(char *expr) {
	int i;
	int j=0;
	for(i=0;expr[i];i++)
		if (expr[i]!=' ') expr[j++]= expr[i];
	expr[j]= 0;
	return i!=j;
}


-[private] int subs(char *dst,char *exp,char *res) {
	char *s;
	int i;
	s= strstr(dst,exp);
	if (!s)
		return 0;
	for(i=0;res[i] && exp[i];i++)
		s[i]= res[i];
	for(;exp[i];i++)
		s[i]= ' ';
	compact(dst);
	return 1;
}

/* ohoho */
-[private] int eval_expr(char *expr) {
	int change;
	do {
		change= 0;
		while(subs(expr,"0&0","0")) change++;
		while(subs(expr,"0&1","0")) change++;
		while(subs(expr,"1&0","0")) change++;
		while(subs(expr,"1&1","1")) change++;
		while(subs(expr,"0|0","0")) change++;
		while(subs(expr,"0|1","1")) change++;
		while(subs(expr,"1|0","1")) change++;
		while(subs(expr,"1|1","1")) change++;
		while(subs(expr,"(0)","0")) change++;
		while(subs(expr,"(1)","1")) change++;
		while(subs(expr,"!0", "1")) change++;
		while(subs(expr,"!1", "0")) change++;
		if (compact(expr)) change++;
		if (expr[1]==0) 
			return expr[0]=='1';
	} while(change);
	return 0;
}

=[ prepare the sink ]
	sink.bullet= ignore_bullet;
	sink.pbreak= ignore_pbreak;
	sink.link= ignore_link;
	sink.text= ignore_text;
	sink.formatted= ignore_formatted;
	sink.header= check_header;
	sink.tag= check_tag;

=[]


=[ evaluate clauses about unknown tags ]
{
	char *begin,*end;
	char tag[200];
	end= query;
	while(get_arg(end,"t!",&begin,&end,tag))
		replace(begin,end,"1");
	end= query;
	while(get_arg(end,"t=",&begin,&end,tag))
		replace(begin,end,"0");
}

=[]

-[unused symbol selist_prepend]
-[unused symbol selist_init]

